from __future__ import with_statement
import os
from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import Qt, SIGNAL, QVariant as QVar
import settings as s
from contextlib import nested

class _Column:
    def __init__(self, visible=True, width=None, weight=None):
        self.width = width
        self.weight = weight
        self.visible = visible
        if width is None and weight is None:
            self.width = 100

class ColumnsWidthSolver:
    WidthRole, WeightRole, VisibleRole = [Qt.UserRole+i for i in xrange(3)]
    def __init__(self, model, min_width=30):
        self._columns = []
        for i in xrange(model.columnCount()):
            role = ColumnsWidthSolver.WidthRole
            width, ok = model.headerData(i, Qt.Horizontal, role).toInt()
            if not ok: width = None
            role = ColumnsWidthSolver.WeightRole
            weight, ok = model.headerData(i, Qt.Horizontal, role).toInt()
            if not ok: weight = None
            role = ColumnsWidthSolver.VisibleRole
            visible = model.headerData(i, Qt.Horizontal, role)
            if visible.type() == QVar.Bool:
                visible = visible.toBool()
            else:
                visible = True
            self._columns.append(_Column(visible, width, weight))
        self._width = None
        self._min_width = min_width
        self._recalc_widths = True
    def count(self):
        return len(self._columns)
    def minimal_width(self):
        min = 0
        for c in self._columns:
            if not c.visible: continue
            if c.weight is not None:
                min += self._min_width
            else:
                min += c.width
        return min
    def is_visible(self, section):
        return self._columns[section].visible
    def set_visible(self, section, param):
        self._columns[section].visible = param
        self._recalc_widths = True
    def is_expandable(self, section):
        return self._columns[section].weight is not None
    def width(self, section):
        self._calc_widths()
        width = self._columns[section].width
        return width
    def weight(self, section):
        return self._columns[section].weight
    def set_width(self, width):
        self._width = width
        self._recalc_widths = True
    def move_border(self, index, delta):
        next_index = index + 1
        while not self.is_visible(next_index): next_index += 1
        self._calc_widths()
        if self._columns[index].width + delta < self._min_width:
            delta = self._min_width - self._columns[index].width
        elif self._columns[next_index].width - delta < self._min_width:
            delta = self._columns[next_index].width - self._min_width
        self._columns[index].width += delta
        self._columns[next_index].width -= delta
        self._calc_weights()
    def reset(self):
        self._recalc_widths = True
    def load_columns_state(self, settings, group=None):
        with nested(s.group(settings, group),
                s.read_array(settings, 'ColumnsState')) as (_, saved_count):
            if saved_count != self.count():
                return
            for i in xrange(saved_count):
                settings.setArrayIndex(i)
                c = self._columns[i]
                if c.weight is None:
                    width, ok = settings.value('width', QVar(c.width)).toInt()
                    c.width = width
                else:
                    weight, ok = settings.value('weight', QVar(c.weight)).toDouble()
                    c.weight = weight
                visible = settings.value('visible', QVar(c.visible)).toBool()
                c.visible = visible
        self.reset()
    def save_columns_state(self, settings, group=None):
        with nested(s.group(settings, group),
                s.write_array(settings, 'ColumnsState', self.count())):
            for i in xrange(self.count()):
                settings.setArrayIndex(i)
                if self.is_expandable(i):
                    weight = self.weight(i)
                    settings.setValue('weight', QVar(weight))
                else:
                    width = self.width(i)
                    settings.setValue('width', QVar(width))
                visible = self.is_visible(i)
                settings.setValue('visible', QVar(visible))
    def _calc_widths(self):
        if self._width is None or not self._recalc_widths: return
        self._recalc_widths = False
        exp_list = []
        const_list = []
        for c in self._columns:
            if not c.visible: continue
            if c.weight is None:
                const_list.append(c)
            else:
                exp_list.append(c)
        while not self.__recalc_widths(exp_list, const_list, self._width):
            pass
    def _calc_weights(self):
        weights_sum = 0.
        widths_sum = 0
        for c in self._columns:
            if not c.visible or c.weight is None: continue
            weights_sum += c.weight
            widths_sum += c.width
        for c in self._columns:
            if not c.visible or c.weight is None: continue
            c.weight = c.width*weights_sum/widths_sum
    @staticmethod
    def __sum(columns):
        ret = 0
        for c in columns:
            ret += c.width
        return ret
    @staticmethod
    def __wsum(columns):
        ret = 0
        for c in columns:
            ret += c.weight
        return ret
    def __recalc_widths(self, elist, clist, width):
        free_space = width - ColumnsWidthSolver.__sum(clist)
        weight_sum = ColumnsWidthSolver.__wsum(elist)
        for c in elist:
            c.width = c.weight*free_space/weight_sum
            if c.width < self._min_width:
                c.width = self._min_width
                clist.append(c)
                elist.remove(c)
                return False
        return True
    def _print_columns(self):
        for c in self._columns:
            print '%s: %s:%s %s' % (c.name, c.width, c.weight, c.visible)


class _ColumnToggleAction(QtGui.QAction):
    def __init__(self, section, table, parent=None):
        name = table.model().headerData(section, Qt.Horizontal, Qt.DisplayRole).toString()
        QtGui.QAction.__init__(self, name, parent)
        self.setCheckable(True)
        checked = table.solver.is_visible(section)
        self.setChecked(checked)
        self._section = section
        self._table = table
        self.connect(self, SIGNAL('toggled(bool)'), self.toggled)
    def toggled(self, checked):
        self._table.solver.set_visible(self._section, checked)
        self._table.set_columns_widths()

class TableView(QtGui.QTableView):
    def __init__(self, parent=None):
        QtGui.QTableView.__init__(self, parent)
        self.setSelectionBehavior(self.SelectRows)
        self.setTabKeyNavigation(False)
        self.verticalHeader().hide()
        self.connect(self.horizontalHeader(), SIGNAL('sectionResized(int,int,int)'), self._columnResized)
        self.horizontalHeader().setContextMenuPolicy(Qt.ActionsContextMenu)
        self.solver = None
    def setModel(self, model):
        if model is not None:   
            self.connect(model, SIGNAL('Exception'), self.__model_exception)
        if self.model() is not None:
            self.disconnect(self.model(), SIGNAL('Exception'), self.__model_exception)
        QtGui.QTableView.setModel(self, model)
        if model is not None:
            s = SIGNAL('selectionChanged(const QItemSelection&,const QItemSelection&)')
            self.connect(self.selectionModel(), s, self.__selectionChanged)
        if model is None:
            self.solver = None
            return
        self.solver = ColumnsWidthSolver(model)
        self.solver.set_width(self.width())
        self.update_columns_state()
    def update_columns_state(self):
        self.set_columns_widths()
        header = self.horizontalHeader()
        for a in header.actions():
            header.removeAction(a)
        for i in xrange(self.solver.count()):
            action = _ColumnToggleAction(i, self, header)
            header.addAction(action)
    def set_columns_widths(self):
        min_w = self.solver.minimal_width()
        min_w += self.width() - self.viewport().width()
        self.setMinimumWidth(min_w)
        self.horizontalHeader().blockSignals(True)
        for i in xrange(self.solver.count()):
            if self.solver.is_visible(i):
                self.setColumnHidden(i, False)
                w = int(self.solver.width(i))
                self.setColumnWidth(i, w)
            else:
                self.setColumnHidden(i, True)
        self.horizontalHeader().blockSignals(False)
    def resizeEvent(self, event):
        if self.model() is None:
            return
        self.solver.set_width(event.size().width())
        self.set_columns_widths()
        QtGui.QTableView.resizeEvent(self, event)
    def load_state(self, settings, group=None):
        if self.solver is not None:
            self.solver.load_columns_state(settings, group)
            self.update_columns_state()
    def save_state(self, settings, group=None):
        if self.solver is not None:
            self.solver.save_columns_state(settings, group)
    def _columnResized(self, column, oldsize, size):
        if self.model() is None:
            return
        self.solver.move_border(column, size-oldsize)
        self.set_columns_widths()
    def __model_exception(self, exc_info):
        self.emit(SIGNAL('Exception'), exc_info)
    def __selectionChanged(self, selected, deselected):
        self.emit(SIGNAL('selectionChanged'), selected, deselected)

