from ice import Communicator
from Tartarus.iface import core
import os

class Cache:
    """Cache class. _recv and _count methods should be
    implemented to use. Receive objects and store it to
    cache member."""
    @staticmethod
    def time():
        return os.times()[4]
    def __init__(self, expire=15):
        self._expire = expire
        self._cache = {}
        self._count_cache = None
    def get(self, index):
        self._clean_cache()
        if index not in self._cache:
            self._recv_in_cache(index)
        return self._cache[index]
    def count(self):
        if (self._count_cache is None or
                Cache.time() - self._count_cache[1] > self._expire):
            count = self._count()
            self._count_cache = count, Cache.time()
        return self._count_cache[0]
    def reset(self):
        self._cache = {}
        self._count_cache = None
    def _recv_in_cache(self, index):
        limit, offset = self._solve_offset(index)
        for obj in self._recv(limit, offset):
            self._push_in_cache(obj, offset)
            offset += 1
    def _solve_offset(self, index):
        for i in xrange(15):
            ind = index + i + 1
            if ind in self._cache:
                break
        end = ind - 1
        for j in xrange(15):
            ind = index - j - 1
            if ind < 0:
                break
            if ind in self._cache:
                break
        begin = ind + 1
        return end - begin, begin
    def _push_in_cache(self, obj, offset):
        obj.cache_push_time = Cache.time()
        self._cache[offset] = obj
    def _clean_cache(self):
        t = Cache.time() - self._expire
        for key, value in self._cache.items():
            if value.cache_push_time < t:
                del self._cache[key]
    def _count(self):
        raise RuntimeError('count method is not implemented in %s class' % self.__class__)
    def _recv(self, limit, offset):
        raise RuntimeError('_recv method is not implemented in %s class' % self.__class__)


class UsersCache(Cache):
    def __init__(self, expire=15):
        Cache.__init__(self, expire)
        self._c = Communicator()
        self._name_cache = {}
        self._id_cache = {}
    def get_by_id(self, id):
        self._clean_cache()
        if id not in self._id_cache:
            u = self._c.um().getById(id)
            self._set_enabled(u)
            self._push_in_cache(u)
        return self._id_cache[id]
    def get_by_name(self, name):
        self._clean_cache()
        if name not in self._name_cache:
            u = self._c.um().getByName(name)
            self._set_enabled(u)
            self._push_in_cache(u)
        return self._name_cache[name]
    def get_by_ids(self, ids):
        self._clean_cache()
        ret = []
        uncached = []
        for uid in ids:
            if uid not in self._id_cache:
                uncached.append(uid)
            else:
                ret.append(self._id_cache[uid])
        users = self._c.um().getUsers(uncached)
        for u in users:
            self._set_enabled(u)
            self._push_in_cache(u)
            ret.append(u)
        return ret
    def search(self, search_str, limit=100):
        ret = []
        for u in self._c.um().search(search_str, limit):
            self._set_enabled(u)
            self._push_in_cache(u)
            ret.append(u)
        return ret
    def reset(self):
        Cache.reset(self)
        self._name_cache = {}
        self._id_cache = {}
    def _set_enabled(self, user_info):
        try:
            user_info.enabled = self._c.kadm().isPrincEnabled(user_info.name)
            user_info.has_password = True
        except core.NotFoundError:
            user_info.enabled = True
            user_info.has_password = False
    def _count(self):
        return self._c.um().count()
    def _recv(self, limit, offset):
        users = self._c.um().get(limit, offset)
        for u in users:
            self._set_enabled(u)
            yield u
    def _push_in_cache(self, user, offset=-1):
        if offset < 0:
            user.cache_push_time = Cache.time()
        else:
            Cache._push_in_cache(self, user, offset)
        self._id_cache[user.uid] = user
        self._name_cache[user.name] = user
    def _clean_cache(self):
        exp_time = Cache.time() - self._expire
        del_keys = []
        for key, value in self._name_cache.iteritems():
            if value.cache_push_time < exp_time:
                del_keys.append(key)
        for key in del_keys:
            del self._name_cache[key]
        del_keys = []
        for key, value in self._id_cache.iteritems():
            if value.cache_push_time < exp_time:
                del_keys.append(key)
        for key in del_keys:
            del self._id_cache[key]
        Cache._clean_cache(self)

_global_users_cache = None

def GlobalUsersCache():
    global _global_users_cache
    if _global_users_cache is None:
        _global_users_cache = UsersCache()
    return _global_users_cache


class GroupsCache(Cache):
    def __init__(self, expire=15):
        Cache.__init__(self, expire)
        self._c = Communicator()
        self._name_cache = {}
        self._id_cache = {}
    def get_by_id(self, id):
        self._clean_cache()
        if id not in self._id_cache:
            g = self._c.gm().getById(id)
            self._push_in_cache(g)
        return self._id_cache[id]
    def get_by_name(self, name):
        self._clean_cache()
        if name not in self._name_cache:
            g = self._c.gm().getByName(name)
            self._push_in_cache(g)
        return self._name_cache[name]
    def get_by_ids(self, ids):
        self._clean_cache()
        ret = []
        uncached = []
        for gid in ids:
            if gid not in self._id_cache:
                uncached.append(gid)
            else:
                ret.append(self._id_cache[gid])
        groups = self._c.gm().getGroups(uncached)
        for g in groups:
            self._push_in_cache(g)
            ret.append(g)
        return ret
    def search(self, search_str, limit=100):
        ret = []
        for g in self._c.gm().search(search_str, limit):
            self._push_in_cache(g)
            ret.append(g)
        return ret
    def reset(self):
        Cache.reset(self)
        self._name_cache = {}
        self._id_cache = {}
    def _count(self):
        return self._c.gm().count()
    def _recv(self, limit, offset):
        return self._c.gm().get(limit, offset)
    def _push_in_cache(self, group, offset=-1):
        if offset < 0:
            group.cache_push_time = Cache.time()
        else:
            Cache._push_in_cache(self, group, offset)
        self._id_cache[group.gid] = group
        self._name_cache[group.name] = group
    def _clean_cache(self):
        exp_time = Cache.time() - self._expire
        del_keys = []
        for key, value in self._name_cache.iteritems():
            if value.cache_push_time < exp_time:
                del_keys.append(key)
        for key in del_keys:
            del self._name_cache[key]
        del_keys = []
        for key, value in self._id_cache.iteritems():
            if value.cache_push_time < exp_time:
                del_keys.append(key)
        for key in del_keys:
            del self._id_cache[key]
        Cache._clean_cache(self)

_global_groups_cache = None

def GlobalGroupsCache():
    global _global_groups_cache
    if _global_groups_cache is None:
        _global_groups_cache = GroupsCache()
    return _global_groups_cache

